#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Parsers for scripts.

File: parsers.py
Author: huxuan
Email: i(at)huxuan.org
"""
import argparse

from slothstock import __version__

# Buy only
BUY_PARSER = argparse.ArgumentParser(add_help=False)
_BUY_GROUP = BUY_PARSER.add_argument_group('Buy')
_BUY_GROUP.add_argument('--reserve-st', action='store_true')
_BUY_GROUP.add_argument('--reserve-suspend', action='store_true')
_BUY_GROUP.add_argument('--check-great-great-grandparent', action='store_true')

# MISC
MISC_PARSER = argparse.ArgumentParser(add_help=False)
MISC_PARSER.add_argument('--daemon', action='store_true')
MISC_PARSER.add_argument('--interval', default=0.1, type=float)
MISC_PARSER.add_argument('--ignore-empty', action='store_true')
MISC_PARSER.add_argument('--output')
MISC_PARSER.add_argument('-V', '--version', action='version',
                         version=__version__)

# SlothStock
SLOTHSTOCK_PARSER = argparse.ArgumentParser(add_help=False)
_SLOTHSTOCK_GROUP = SLOTHSTOCK_PARSER.add_argument_group('SlothStock')
_SLOTHSTOCK_GROUP.add_argument('--ebk', action='append', default=[])
_SLOTHSTOCK_GROUP.add_argument('--period', default='day')
_SLOTHSTOCK_GROUP.add_argument('--loose', action='store_true')
_SLOTHSTOCK_GROUP.add_argument('--skip-child', action='store_true')

# WxPusher
WXPUSHER_PARSER = argparse.ArgumentParser(add_help=False)
_WXPUSHER_GROUP = WXPUSHER_PARSER.add_argument_group('WxPusher')
_WXPUSHER_GROUP.add_argument('--title')
_WXPUSHER_GROUP.add_argument('--token')
_WXPUSHER_GROUP.add_argument('--topic-ids', action="append", default=[])
_WXPUSHER_GROUP.add_argument('--uids', action='append', default=[])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Collection of provider modules.

File: __init__.py
Author: huxuan
Email: i(at)huxuan.org
"""
from slothstock.providers.xueqiu import XueQiu

__all__ = [
    'XueQiu',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utilities for Slothstock.

File: utils.py
Author: huxuan
Email: i(at)huxuan.org
"""
from faker import Faker
import numpy
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from slothstock import constants


def export_ebk(symbols, filename='sloth.ebk'):
    """Export symbols in ebk format."""
    res = [b'']
    for symbol in symbols:
        code, exchange = symbol.split('.')
        if exchange == 'SH':
            res.append(f'1{code}'.encode())
        else:
            res.append(f'0{code}'.encode())

    with open(filename, 'wb') as fout:
        fout.write(b'\r\n'.join(res))


def import_ebk(filename):
    """Import symbols from ebk format."""
    res = []
    with open(filename, 'rb') as fin:
        fin.readline()  # The first line is useless.
        for line in fin:
            line = line.decode().strip()
            if line[0] == '1':
                res.append(f'{line[1:]}.SH')
            else:
                res.append(f'{line[1:]}.SZ')
    return res


def is_st(stock):
    """Check whether it is ST."""
    return 'ST' in stock.get('name')


def is_suspend(stock):
    """Check whether it is suspended."""
    return stock.get('type') == 11 and numpy.isnan(stock.get('amplitude'))


def create_fake_session(index_url=None):
    """Create fake seesion."""
    session = requests.Session()
    session.headers[constants.HEADER_USER_AGENT] = Faker().internet_explorer()
    if index_url:  # Browser index page for cookies.
        requests_retry_session(session).get(index_url)
    return session


def requests_retry_session(
        session=None,
        retries=3,
        backoff_factor=0.3,
        status_forcelist=(500, 502, 504)):
    """Send requests with retry."""
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Xueqiu provider.

File: xueqiu.py
Author: huxuan
Email: i(at)huxuan.org
"""
from datetime import datetime
from datetime import timedelta

import pandas

from slothstock import constants
from slothstock import exceptions
from slothstock import utils

XUEQIU_ETF_LIST_URL = 'https://xueqiu.com/service/v5/stock/screener/fund/list'
XUEQIU_KLINE_URL = 'https://stock.xueqiu.com/v5/stock/chart/kline.json'
XUEQIU_LIST_URL = 'https://xueqiu.com/service/v5/stock/screener/quote/list'
XUEQIU_QUOTE_URL = 'https://xueqiu.com/service/v5/stock/batch/quote'
XUEQIU_URL = 'https://xueqiu.com/'


def symbol_transform(symbol):
    """Transform symbol for XueQiu."""
    return ''.join(symbol.split('.')[::-1])


def symbol_restore(symbol):
    """Restore symbol for XueQiu."""
    return f'{symbol[2:]}.{symbol[:2]}'


def datetime_to_timestamp(dt_origin=None):
    """Convert local datetime to utc timestamp."""
    if dt_origin is None:
        dt_origin = datetime.utcnow() + timedelta(hours=8)
    elif isinstance(dt_origin, str):
        dt_origin = datetime.strptime(dt_origin, constants.DATETIME_FORMAT)
    elif not isinstance(dt_origin, datetime):
        raise exceptions.InvalidDatetimeError(dt_origin)
    return int(dt_origin.timestamp() * 1000)


def timestamp_to_datetime(timestamp):
    """Convert utc timestamp to local datetime."""
    return datetime.fromtimestamp(timestamp / 1000)


def check_response(res):
    """Roughly check result in common."""
    if not res:
        raise exceptions.NoResponseError(res.status_code)

    try:
        res = res.json()
    except ValueError:
        return exceptions.InvalidResultError(res.content)

    if res.get('error_code') != 0:
        return exceptions.XueQiuError(
            res['error_code'],
            res['error_description'])

    if not res.get('data'):
        raise exceptions.NoResultError(res.content)

    return res


class MetaXueQiu(type):
    """Metaclass for XueQiu."""

    session = utils.create_fake_session(XUEQIU_URL)

    @classmethod
    def kline(cls, symbol, period='day', begin=None, count=142):
        """Fetch kline information."""
        params = {
            'begin': datetime_to_timestamp(begin),
            'count': -abs(count),
            'indicator': 'kline',
            'period': period,
            'symbol': symbol_transform(symbol),
            'type': 'before',
        }
        res = utils.requests_retry_session(cls.session).get(
            XUEQIU_KLINE_URL, params=params)
        res = check_response(res)

        data = pandas.DataFrame(
            data=res['data'].get('item'),
            columns=res['data'].get('column'))
        data.rename(columns={'timestamp': 'datetime'}, inplace=True)
        data.datetime = data.datetime.apply(timestamp_to_datetime)
        data.set_index('datetime', inplace=True)
        return data

    @classmethod
    def list(cls, url=XUEQIU_LIST_URL, **params):
        """List stock information."""
        params.update({
            '_': datetime_to_timestamp(),
            'order': params.get('reverse', False) and 'desc' or 'asc',
            'order_by': params.get('order_by', 'symbol'),
            'page': params.get('page', 1),
            'size': params.get('size', 4000),
            'type': params.get('type', 'sh_sz'),
        })
        res = utils.requests_retry_session(cls.session).get(url, params=params)
        res = check_response(res)

        data = pandas.DataFrame(res['data']['list'])
        data.symbol = data.symbol.apply(symbol_restore)
        data.set_index('symbol', inplace=True)
        return data

    @classmethod
    def list_ashare(cls, **params):
        """List A Share stocks."""
        data = cls.list(**params)
        return data[data.index.str[:2].isin(constants.STOCK_PREFIX_MARKET)]

    @classmethod
    def list_etf(cls, **params):
        """List ETF."""
        params.update({
            'parent_type': 1,
            'size': params.get('size', 300),
            'type': 18,
        })
        return cls.list(XUEQIU_ETF_LIST_URL, **params)

    @classmethod
    def list_index(cls, indexes=None):
        """List Index."""
        url = XUEQIU_QUOTE_URL
        indexes = indexes or [symbol_transform(index)
                              for index in constants.STOCK_INDEX]
        params = {
            '_': datetime_to_timestamp(),
            'symbol': ','.join(indexes or constants.STOCK_INDEX),
        }
        res = utils.requests_retry_session(cls.session).get(url, params=params)
        res = check_response(res)

        data = pandas.DataFrame([
            item['quote'] for item in res['data'].get('items', [])
        ])
        data.symbol = data.symbol.apply(symbol_restore)
        data.set_index('symbol', inplace=True)
        return data

    @classmethod
    def list_stocks(cls, ebk=None):
        """List stocks."""
        stocks_ashare = XueQiu.list_ashare()
        stocks_etf = XueQiu.list_etf()
        stocks_index = XueQiu.list_index()
        stocks = pandas.concat([stocks_ashare, stocks_etf, stocks_index])

        if ebk:
            res = set()
            for filename in ebk:
                res.update(utils.import_ebk(filename))
            stocks = stocks[stocks.index.isin(res)]

        return stocks


class XueQiu(metaclass=MetaXueQiu):  # pylint: disable=too-few-public-methods
    """XueQiu provider."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Common utilities for scripts.

File: common.py
Author: huxuan
Email: i(at)huxuan.org
"""
from datetime import datetime
from datetime import timedelta

from wxpusher import WxPusher

from slothstock import utils


def combine_content(title, stocks):
    """Combine title and stocks to message content."""
    content = [title]
    content.extend([
        f'{symbol} {stocks.loc[symbol, "name"]}'
        for symbol in stocks.index
    ])
    content.append(str(datetime.utcnow() + timedelta(hours=8)))
    return '\n'.join(content)


def show_result(stocks, action, args):
    """Show results."""
    # Return for empty result when set to ignore empty.
    if args.ignore_empty and stocks.index.empty:
        return

    # Output a file when output is set.
    if args.output:
        utils.export_ebk(stocks.index, args.output)

    # Send notification if token and (uids or topic_ids) is set.
    if args.token and (args.uids or args.topic_ids):
        # Format title.
        title = args.title
        if not title:
            if stocks.index.empty:
                title = f'{args.period}级别无可能{action}点'
            else:
                title = f'{args.period}级别可能{action}点'

        content = combine_content(title, stocks)
        WxPusher.send_message(content,
                              token=args.token,
                              uids=args.uids,
                              topic_ids=args.topic_ids)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unittests for utils.

File: test_utils.py
Author: huxuan
Email: i(at)huxuan.org
"""
import random
import tempfile
import unittest

from slothstock import utils
from slothstock.providers import XueQiu


class TestUtils(unittest.TestCase):
    """Unittests for utils."""

    @classmethod
    def setUpClass(cls):
        """Set up for class."""
        cls.stocks = XueQiu.list_stocks()

    def test_import_export_ebk(self):
        """Positive case for import & export ebk."""
        temp_file = tempfile.NamedTemporaryFile()
        stocks = XueQiu.list_ashare(
            page=random.randint(1, 20),
            size=random.randint(1, 100))
        utils.export_ebk(stocks.index, temp_file.name)
        symbols = utils.import_ebk(temp_file.name)
        for symbol in symbols:
            self.assertIn(symbol, stocks.index)

    def test_is_st(self):
        """Cases to check whether is ST."""
        self.assertTrue(utils.is_st(self.stocks.loc['600518.SH']))
        self.assertFalse(utils.is_st(self.stocks.loc['000001.SZ']))

    def test_is_suspend(self):
        """Cases to check whether is suspended."""
        self.assertTrue(utils.is_suspend(self.stocks.loc['000029.SZ']))
        self.assertTrue(utils.is_suspend(self.stocks.loc['002450.SZ']))
        self.assertTrue(utils.is_suspend(self.stocks.loc['300104.SZ']))
        self.assertTrue(utils.is_suspend(self.stocks.loc['600145.SH']))

        self.assertFalse(utils.is_suspend(self.stocks.loc['000001.SH']))
        self.assertFalse(utils.is_suspend(self.stocks.loc['510050.SH']))
        self.assertFalse(utils.is_suspend(self.stocks.loc['600519.SH']))

        self.assertFalse(utils.is_suspend(self.stocks.loc['000001.SZ']))
        self.assertFalse(utils.is_suspend(self.stocks.loc['159949.SZ']))
        self.assertFalse(utils.is_suspend(self.stocks.loc['399001.SZ']))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unittests for MACD.

File: test_macd_indicator.py
Author: huxuan
Email: i(at)huxuan.org
"""
from datetime import datetime
import unittest

from slothstock.indicators.macd_indicator import clean_macd
from slothstock.indicators.macd_indicator import is_about_to_bottom_divergence
from slothstock.indicators.macd_indicator import is_about_to_death_cross
from slothstock.indicators.macd_indicator import is_about_to_golden_cross
from slothstock.indicators.macd_indicator import is_about_to_top_divergence
from slothstock.indicators.macd_indicator import is_death_cross
from slothstock.indicators.macd_indicator import is_golden_cross
from slothstock.indicators.macd_indicator import is_negative
from slothstock.indicators.macd_indicator import is_positive
from slothstock.providers import XueQiu


class TestMACD(unittest.TestCase):
    """Unittests for MACD."""

    def test_is_about_to_top_divergence(self):
        """Positive case for top divergence."""
        res = XueQiu.kline('000016.SH', 'day', datetime(2019, 10, 21))
        macd, macdsignal, macdhist = clean_macd(res.close)
        self.assertTrue(is_about_to_top_divergence(
            macd, macdsignal, macdhist))

    def test_is_about_to_bottom_divergence(self):
        """Positive case for bottom divergence."""
        res = XueQiu.kline('000016.SH', 'week', datetime(2019, 1, 11))
        macd, macdsignal, macdhist = clean_macd(res.close)
        self.assertTrue(is_about_to_bottom_divergence(
            macd, macdsignal, macdhist))

    def test_is_negative(self):
        """Test is_negative() with loose."""
        res = XueQiu.kline('000001.SH', 'day', datetime(2020, 2, 6))
        macd, macdsignal, _ = clean_macd(res.close)
        self.assertTrue(is_negative(macd, macdsignal))
        self.assertTrue(is_negative(macd, macdsignal, True))
        self.assertFalse(is_negative(macd[:-1], macdsignal[:-1]))
        self.assertTrue(is_negative(macd[:-1], macdsignal[:-1], True))

    def test_is_positive(self):
        """Test is_positive() with loose."""
        res = XueQiu.kline('000001.SH', 'day', datetime(2019, 12, 18))
        macd, macdsignal, _ = clean_macd(res.close)
        self.assertTrue(is_positive(macd, macdsignal))
        self.assertTrue(is_positive(macd, macdsignal, True))
        self.assertFalse(is_positive(macd[:-1], macdsignal[:-1]))
        self.assertTrue(is_positive(macd[:-1], macdsignal[:-1], True))

    def test_is_death_cross(self):
        """Test is_death_cross() with loose."""
        res = XueQiu.kline('000001.SH', 'day', datetime(2020, 2, 6))
        _, _, macdhist = clean_macd(res.close)
        self.assertFalse(is_death_cross(macdhist))
        self.assertTrue(is_death_cross(macdhist, True))
        self.assertTrue(is_death_cross(macdhist[:-1]))
        self.assertTrue(is_death_cross(macdhist[:-1], True))

    def test_is_golden_cross(self):
        """Test is_golden_cross() with loose."""
        res = XueQiu.kline('000001.SH', 'day', '2019-12-19 00:00')
        _, _, macdhist = clean_macd(res.close)
        self.assertFalse(is_golden_cross(macdhist))
        self.assertTrue(is_golden_cross(macdhist, True))
        self.assertTrue(is_golden_cross(macdhist[:-1]))
        self.assertTrue(is_golden_cross(macdhist[:-1], True))

    def test_is_about_death_cross(self):
        """Test is_about_to_death_cross() with loose."""
        res = XueQiu.kline('000001.SH', 'day', datetime(2019, 12, 19))
        _, _, macdhist = clean_macd(res.close)
        self.assertTrue(is_about_to_death_cross(macdhist))
        self.assertTrue(is_about_to_death_cross(macdhist, True))
        self.assertFalse(is_about_to_death_cross(macdhist[:-1]))
        self.assertTrue(is_about_to_death_cross(macdhist[:-1], True))

    def test_is_about_golden_cross(self):
        """Test is_about_to_golden_cross() with loose."""
        res = XueQiu.kline('000001.SH', 'day', datetime(2020, 2, 6))
        _, _, macdhist = clean_macd(res.close)
        self.assertTrue(is_about_to_golden_cross(macdhist))
        self.assertTrue(is_about_to_golden_cross(macdhist, True))
        self.assertFalse(is_about_to_golden_cross(macdhist[:-1]))
        self.assertTrue(is_about_to_golden_cross(macdhist[:-1], True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unittests for XueQiu.

File: test_xueqiu.py
Author: huxuan
Email: i(at)huxuan.org
"""
import random
import unittest

import pandas

from slothstock.providers import XueQiu


class TestXueQiu(unittest.TestCase):
    """Unittests for XueQiu."""

    def test_kline(self):
        """Positive case for kline."""
        stocks = XueQiu.list_ashare(page=random.randint(1, 1000), size=1)
        res = XueQiu.kline(stocks.index[0])
        self.assertIsInstance(res, pandas.DataFrame)
        self.assertGreater(len(res.index), 0)

    def test_list(self):
        """Positive case for list."""
        res = XueQiu.list()
        self.assertIsInstance(res, pandas.DataFrame)
        self.assertGreater(len(res.index), 0)

    def test_list_ashare(self):
        """Positive case for list_ashare."""
        res = XueQiu.list_ashare()
        self.assertIsInstance(res, pandas.DataFrame)
        self.assertGreater(len(res.index), 0)

    def test_list_etf(self):
        """Positive case for list_etf."""
        res = XueQiu.list_etf()
        self.assertIsInstance(res, pandas.DataFrame)
        self.assertGreater(len(res.index), 0)

    def test_list_index(self):
        """Positive case for list_index."""
        res = XueQiu.list_index()
        self.assertIsInstance(res, pandas.DataFrame)
        self.assertEqual(len(res.index), 4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom exceptions.

File: exceptions.py
Author: huxuan
Email: i(at)huxuan.org
"""


class BaseCustomException(RuntimeError):
    """Base Custom Exception."""


class InvalidDatetimeError(BaseCustomException):
    """Raised for invalid datetime."""


class InvalidResultError(BaseCustomException):
    """Raised when the result can not be parsed."""


class NoResponseError(BaseCustomException):
    """Raised when no response fetched."""


class NoResultError(BaseCustomException):
    """Raised when no result fetched."""


class XueQiuError(BaseCustomException):
    """Raised for specific error from XueQiu."""

    def __init__(self, code, message):
        """Init for XueQiuError."""
        super(XueQiuError, self).__init__()
        self.code = code
        self.message = message

    def __str__(self):
        """Str for XueQiuError."""
        return f'[{self.code}] {self.message}'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Script indicating buy signals.

File: sloth_buy.py
Author: huxuan
Email: i(at)huxuan.org
"""
import argparse
import time

from tqdm import tqdm

from slothstock import constants
from slothstock import utils
from slothstock.indicators import macd_indicator
from slothstock.providers import XueQiu
from slothstock.scripts import common
from slothstock.scripts import parsers


def parse_args():
    """Argument Parser."""
    parser = argparse.ArgumentParser(parents=[
        parsers.BUY_PARSER,
        parsers.MISC_PARSER,
        parsers.SLOTHSTOCK_PARSER,
        parsers.WXPUSHER_PARSER,
    ])
    return parser.parse_args()


def check_buy(stocks, args):
    """Check buy signals."""
    symbols = stocks.sample(frac=1).index
    idx = constants.PERIODS.index(args.period)
    candidate_buy = set()

    pbar = tqdm(symbols, disable=args.daemon)
    for symbol in pbar:
        pbar.set_description(symbol)

        # Whether to reserve ST stocks.
        if not args.reserve_st and utils.is_st(stocks.loc[symbol]):
            continue

        # Whether to reserve suspended stocks.
        if not args.reserve_suspend and utils.is_suspend(stocks.loc[symbol]):
            continue

        # Check great-great-grandparent period.
        if idx + 4 < len(constants.PERIODS) and \
                args.check_great_great_grandparent:
            time.sleep(args.interval)
            period_cur = constants.PERIODS[idx + 4]
            kline = XueQiu.kline(symbol, period_cur)
            _, _, macdhist = macd_indicator.clean_macd(kline.close)
            if not macd_indicator.is_golden_cross(macdhist, args.loose):
                continue

        # Check grandparent period.
        if idx + 2 < len(constants.PERIODS):
            time.sleep(args.interval)
            period_cur = constants.PERIODS[idx + 2]
            kline = XueQiu.kline(symbol, period_cur)
            _, _, macdhist = macd_indicator.clean_macd(kline.close)
            if not macd_indicator.is_golden_cross(macdhist, args.loose):
                continue

        # Check current period.
        time.sleep(args.interval)
        kline = XueQiu.kline(symbol, args.period)
        macd, macdsignal, macdhist = macd_indicator.clean_macd(kline.close)
        if not macd_indicator.is_negative(macd, macdsignal, args.loose):
            continue
        if not macd_indicator.is_about_to_golden_cross(macdhist, args.loose):
            continue

        # Check child period.
        if idx > 0 and not args.skip_child:
            time.sleep(args.interval)
            period_cur = constants.PERIODS[idx - 1]
            kline = XueQiu.kline(symbol, period_cur)
            macd, macdsignal, macdhist = macd_indicator.clean_macd(kline.close)
            if not macd_indicator.is_about_to_bottom_divergence(
                    macd, macdsignal, macd_indicator, args.loose):
                continue

        candidate_buy.add(symbol)
    return stocks[stocks.index.isin(candidate_buy)].sort_index()


def main():
    """Indicate buy signals."""
    args = parse_args()

    stocks = XueQiu.list_stocks(args.ebk)
    stocks = check_buy(stocks, args)
    common.show_result(stocks, '买', args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
MACD indicator.

File: macd.py
Author: huxuan
Email: i(at)huxuan.org
"""
import numpy
import talib


def clean_macd(closes, fastperiod=12, slowperiod=26, signalperiod=9):
    """Calculate MACD with clean up."""
    macd, macdsignal, macdhist = talib.MACD(
        closes, fastperiod, slowperiod, signalperiod)
    macd = macd[~numpy.isnan(macd)]
    macdsignal = macdsignal[~numpy.isnan(macdsignal)]
    macdhist = macdhist[~numpy.isnan(macdhist)]
    return macd, macdsignal, macdhist


def is_negative(macd, macdsignal, loose=False):
    """Check whether the macd is negative."""
    if len(macd) < 1:
        return False
    if not loose:
        return macd[-1] <= 0 and macdsignal[-1] <= 0
    return macd[-1] <= 0 or macdsignal[-1] <= 0


def is_positive(macd, macdsignal, loose=False):
    """Check whether the macd is positive."""
    if len(macd) < 1:
        return False
    if not loose:
        return macd[-1] >= 0 and macdsignal[-1] >= 0
    return macd[-1] >= 0 or macdsignal[-1] >= 0


def is_death_cross(macdhist, loose=False):
    """Check whether the macd is death cross."""
    if len(macdhist) < 1 or macdhist[-1] > 0:
        return False
    if not loose:
        return len(macdhist) > 1 and macdhist[-2] >= macdhist[-1]
    return True


def is_golden_cross(macdhist, loose=False):
    """Check whether the macd is golden cross."""
    if len(macdhist) < 1 or macdhist[-1] < 0:
        return False
    if not loose:
        return len(macdhist) > 1 and macdhist[-2] <= macdhist[-1]
    return True


def is_about_to_death_cross(macdhist, loose=False):
    """Check whether is going to be death cross."""
    if len(macdhist) < 1 or macdhist[-1] < 0:
        return False
    if not loose:
        return len(macdhist) > 1 and macdhist[-2] >= macdhist[-1]
    return True


def is_about_to_golden_cross(macdhist, loose=False):
    """Check whether is going to be golden cross."""
    if len(macdhist) < 1 or macdhist[-1] > 0:
        return False
    if not loose:
        return len(macdhist) > 1 and macdhist[-2] <= macdhist[-1]
    return True


def is_about_to_top_divergence(macd, macdsignal, macdhist, loose=False):
    """Check whether is going to be top divergence."""
    if not is_positive(macd, macdsignal, loose) or \
            not is_about_to_death_cross(macdhist, loose):
        return False

    idx_negative = numpy.where(macdhist <= 0)[0]
    if len(idx_negative) < 1:
        return False
    idx_golden_cross = idx_negative[-1]

    idx_positive = numpy.where(macdhist[:idx_golden_cross] >= 0)[0]
    if len(idx_positive) < 1:
        return False
    idx_death_cross = idx_positive[-1]

    if not loose and any(macd[idx_death_cross:] < 0):
        return False
    if any(macdsignal[idx_death_cross:] < 0):
        return False
    return macdsignal[idx_death_cross] >= macdsignal[-1]


def is_about_to_bottom_divergence(macd, macdsignal, macdhist, loose=False):
    """Check whether is going to be bottom divergence."""
    if not is_negative(macd, macdsignal, loose) or \
            not is_about_to_golden_cross(macdhist, loose):
        return False

    idx_positive = numpy.where(macdhist >= 0)[0]
    if len(idx_positive) < 1:
        return False
    idx_death_cross = idx_positive[-1]

    idx_negative = numpy.where(macdhist[:idx_death_cross] <= 0)[0]
    if len(idx_negative) < 1:
        return False
    idx_golden_cross = idx_negative[-1]

    if not loose and any(macd[idx_golden_cross:] > 0):
        return False
    if any(macdsignal[idx_golden_cross:] > 0):
        return False
    return macdsignal[idx_golden_cross] <= macdsignal[-1]
